class Errors(list):
    """ A list object which holds a dictionary of errors. And a status code """

    def __init__(self, status=400):
        self.status = status
        super(Errors, self).__init__()

    def add(self, location, desc):
        error = {'location': location,
                 'description': desc}
        self.append(error)



from pyramid.httpexceptions import (HTTPMethodNotAllowed,
                                    HTTPError,
                                    HTTPUnsupportedMediaType,
                                    HTTPNotAcceptable)
from pyramid.response import Response
from pyramid.exceptions import PredicateMismatch
from errors import Errors
from lovely.pyrest.validation import validate_schema
from lovely.pyrest.settings import (
    DEFAULT_VIEW_SETTINGS,
    DEFAULT_HELP_MESSAGE,
    get_jsonp_param_name
)
import json
import functools


def decorate_view(view, args):
    """ Adds a decorator to a `pyramid` view.
    The decorator validates the `schema` and executes the additional
    `validators`.

    :param view: the view to decorate
    :param args: the args to use for the decoration
    """
    def get_help_message(request):
        """ Checks if 'help' is set as query parameter in a request.
        If so, the schema will be displayed as a documentation
        to show which query parameters can (or must) be present in a reqeust.

        If and API endpoint does not accept any query parameters
        (i.e. schema is empty) an appropriate message will be returned.
        """
        schema = args.get('schema')
        response = None
        if 'help' in request.GET:
            if schema:
                response = schema
            else:
                response = DEFAULT_HELP_MESSAGE
        return response

    def wrapper(request):
        help = args.get('help', DEFAULT_VIEW_SETTINGS.get('help'))
        if help:
            response = get_help_message(request)
            # do not validate anything else. show the help imediately.
            if response is not None:
                return response

        _view = view
        validators = args.get('validators', ())
        schema = args.get('schema')
        # check schema
        if schema:
            validate_schema(request, schema)
        # execute validators
        for validator in validators:
            validator(request)
        # the validators or validate_schema() adds errors to the request if
        # validaton fails.
        # check if there are errors and return an ErrorResponse if so
        if len(request.errors) > 0:
            if is_jsonp_request(request):
                # if JSONP is requested return a HTTP 200 OK and create an
                # error message. Pyramid wraps the message with given
                # JSONP function name.
                response = create_json_errors(request.errors)
                response['http_status'] = request.errors.status
                return response
            else:
                return JSONError(request.errors, request.errors.status)
        response = _view(request)
        return response

    functools.wraps(wrapper)
    return wrapper

def is_jsonp_request(request):
    return get_jsonp_param_name() in request.GET


def create_json_errors(errors):
    errors = {
        'status': 'error',
        'errors': errors
    }
    return errors


class JSONError(HTTPError):

    def __init__(self, errors, status=400):
        body = create_json_errors(errors)
        # HTTPError is derived from Response
        super(Response, self).__init__(json.dumps(body))
        self.status = status
        self.content_type = 'application/json'


def get_fallback_view(service):
    """ The fallback_view doesn't have any predicates, except the service route.
        If there is no other view in the `service` which matches the requests
        predicates, pyramid will use this fallback view because it will
        always match the predicates of the request.
        Instead of raising the PredicateMismatch error which would cause a 404
        response, the view checks the `content_type` and `access_header`
        predicate. So it's possible to return a correct error instead of 404.
    """

    # the fallback view returns a correct error at predicate
    # mismatch
    def _fallback(request):
        if request.method not in service.methods:
            errors = Errors()
            errors.add(
                'method',
                'The method %s is not allowed for this resource' % request.method
            )
            return JSONError(errors, HTTPMethodNotAllowed.code)
        # Check other predicates like content type
        for method, _, args in service.definitions:
            if method != request.method:
                continue
            # Check Accept header
            if 'accept' in args:
                supported = service.acceptables(request.method)
                if not request.accept.best_match(supported):
                    request.errors.status = HTTPNotAcceptable.code
                    request.errors.add(
                        'header',
                        'Accept header should be one of %s' % supported
                    )
                    return JSONError(request.errors, request.errors.status)
            # Check Content-Type header
            if 'content_type' in args:
                supported = service.content_types(request.method)
                if request.content_type not in supported:
                    request.errors.status = HTTPUnsupportedMediaType.code
                    request.errors.add(
                        'header',
                        'Content-Type header should be one of %s' % supported)
                    return JSONError(request.errors, request.errors.status)
        raise PredicateMismatch(service.name)

    return _fallback



from docutils.parsers.rst import Directive
from importlib import import_module
from lovely.pyrest.service import get_service
from lovely.pyrest.sphinx.helpers import create_node, trim, empty_node
from docutils import nodes


class ValidatorsDirective(Directive):
    """ The Validators directive renders the validators docstrings """

    has_content = True
    required_arguments = 3

    def run(self):
        module, svc, method = self.arguments
        import_module(module)
        service = get_service(svc)
        service_id = "service_%d" % self.state.document.settings.env.new_serialno('service')
        return [ValidatorsDirective.render(service,
                                           method,
                                           service_id)]

    @staticmethod
    def render(service, method, service_id):
        """ Renders the validators list of the service """
        if not service:
            return empty_node()
        validators = service.get_argument('validators', method)
        if not validators:
            return empty_node()

        rendered_nodes = []
        # render every validator and add it to the `rendered_nodes`
        # the result of rendering is not none
        for validator in validators:
            node = ValidatorsDirective.render_validator(validator)
            if node is not None:
                rendered_nodes.append(node)

        if len(rendered_nodes) > 0:
            validators_id = "service_%s_%s_%s" % (service_id,
                                                  method,
                                                  'validators')
            validator_section = nodes.section(ids=[validators_id])
            validator_section += nodes.title(text='Validation:')
            # add the rendered_nodes to the validator_title
            for node in rendered_nodes:
                validator_section += node
            return validator_section
        # Return an empty node if none of the validators has a docstring
        return empty_node()

    @staticmethod
    def render_validator(validator):
        doc = validator.__doc__
        if doc is not None:
            doc = trim(doc)
            return create_node(doc)
        return None



from lovely.pyrest.service import Service

articles = Service(name='article', path='/article/{id}')


def article_exists(request):
    """ Validates that the article exists """
    pass


@articles.get(validators=[article_exists], accept='application/json')
def get(request):
    """ Returns an article """
    return {'title': 'the title'}


# Create a schema to validate the body
schema = {
    "body": {
        "type": "object",
        "properties": {
            "title": {
                "type": "string"
            }
        }
    },
    "query": {
        "type": "object",
        "properites": {
            "id": {
                "type": "string"
            }
        }
    }
}


@articles.post(schema=schema, content_type='application/json')
def post(request):
    """ Save an article """
    return {'created': '123'}



import venusian
import functools
from lovely.pyrest.settings import DEFAULT_VIEW_SETTINGS as DEFAULTS


METHODS = ['GET', 'POST', 'PUT', 'DELETE', 'HEAD', 'PURGE']

SERVICES = {}


def get_services():
    return SERVICES.values()


def get_service(name):
    """ Returns the service """
    for s in SERVICES.values():
        if s.name == name:
            return s
    return None


class Service(object):

    def __init__(self, name, path, description=None, depth=1):
        self.name = name
        self.path = path
        self.description = description
        self.definitions = []
        self.methods = []

        SERVICES[self.name] = self

        # this callback gets called when config.scan() gets triggered
        # and registers the service itself
        def callback(context, name, ob):
            config = context.config.with_package(info.module)
            config.add_service(self)
        # attach the callback to venusian so it gets called at scan
        info = venusian.attach(self, callback, category='pyramid', depth=depth)

        # add aliases for the decorators
        for m in METHODS:
            setattr(self, m.lower(),
                    functools.partial(self.decorator, m))

    # The decorator to add a view to a service
    def decorator(self, method, **kwargs):
        def wrapper(view):
            self.add_view(method, view, **kwargs)
            return view
        return wrapper

    def add_view(self, method, view, **kwargs):
        method = method.upper()
        args = self.get_args(kwargs)

        self.definitions.append((method, view, args))
        if method not in self.methods:
            self.methods.append(method)

    def content_types(self, method):
        types = []
        for m, v, args in self.definitions:
            if m == method and 'content_type' in args:
                types.append(args['content_type'])
        return types

    def acceptables(self, method):
        acceptable = []
        for m, v, args in self.definitions:
            if m == method and 'accept' in args:
                acceptable.append(args['accept'])
        return acceptable

    def get_args(self, args=None):
        if args is None:
            args = {}
        arguments = {}

        for arg in DEFAULTS:
            # get the value from the passed conf
            # then from the DEFAULTS
            arguments[arg] = args.get(arg, DEFAULTS[arg])
        # update the defaults with the passed args
        arguments.update(args)
        return arguments

    def get_argument(self, key, method):
        """ Checks if any of the defined views with the method
            has the argument set and returns it.
        """
        for m, view, args in self.definitions:
            if m == method.upper():
                return args.get(key)
        return None



import unittest
import doctest
import requests
import sphinx.cmdline
import tempfile
import shutil
import os
from lovely.pyrest.tests import print_json
from contextlib import contextmanager
from webtest import TestApp
from pyramid import testing
from pprint import pprint
from lovely.pyrest.service import SERVICES

DOC_CONF = """
extensions = ['lovely.pyrest.sphinx']
source_suffix = '.txt'
master_doc = 'index'
html_theme = 'pyramid'
"""


@contextmanager
def _temp_dir():
    tmp = tempfile.mkdtemp()
    yield tmp
    shutil.rmtree(tmp)


def render_doc(rst_path):
    """ Renders a restructured Text file and returns the
        html output """

    # Remove all existing services
    SERVICES.clear()
    with _temp_dir() as tmp:
        out_dir = os.path.join(tmp, 'out')
        in_dir = os.path.join(tmp, 'in')
        os.mkdir(in_dir)
        # write the rst string to a file
        conf = DOC_CONF % dict(
            rst_file=rst_path
        )
        shutil.copyfile(rst_path, os.path.join(in_dir, 'index.txt'))
        conf_path = os.path.join(in_dir, 'conf.py')
        with file(conf_path, 'wb') as cf:
            cf.write(conf)
        sphinx_argv = ['sphinx-build', '-b', 'html',
                       in_dir, out_dir]
        sphinx.cmdline.main(sphinx_argv)
        index_path = os.path.join(out_dir, 'index.html')
        with open(index_path, 'r') as index:
            return index.read()


def get_app(module):
    # scans the defined module and returns
    # a TestApp

    # Settings can also be defined in a .ini file.
    settings = {
        'lovely.pyrest.jsonp.param_name': 'callback'
    }
    config = testing.setUp(settings=settings)
    config.include('lovely.pyrest')
    config.scan(module)
    return TestApp(config.make_wsgi_app())


def setUp(test):
    test.globs['requests'] = requests
    test.globs['print_json'] = print_json
    test.globs['get_app'] = get_app
    test.globs['pprint'] = pprint
    test.globs['render_doc'] = render_doc
    pass


def tearDown(test):
    pass


def create_suite(testfile, layer=None, level=None,
                 tearDown=tearDown, setUp=setUp, cls=doctest.DocFileSuite):
    suite = cls(
        testfile, tearDown=tearDown, setUp=setUp,
        optionflags=doctest.NORMALIZE_WHITESPACE | doctest.ELLIPSIS)
    if layer:
        suite.layer = layer
    if level:
        suite.level = level
    return suite


def test_suite():
    s = unittest.TestSuite((
        create_suite('../docs/index.txt'),
        create_suite('../docs/service.txt'),
        create_suite('../docs/validation.txt'),
        create_suite('../docs/sphinx.txt'),
        create_suite('../docs/jsonp.txt'),
        create_suite('../docs/help.txt'),
        create_suite('../docs/accept_catch_all.txt')
    ))
    return s



from lovely.pyrest.service import Service

svc = Service(name='svc', path='/svc')


@svc.get(accept="text/plain", accept_catch_all=False)
def get(request):
    return "'title': 'Hello World!'"

@svc.get(accept='application/json')
def get_json(request):
    return {'title': 'Hello World!'}



from lovely.pyrest.service import Service

"""
This module contains a service definitions and a couple of views to explain
the 'accept_catch_all' view annotation parameter.
"""

service1 = Service('myService1', '/service/1')

@service1.get()
def svc1_get(request):
    return {'test': 'succeeded'}


service2 = Service('myService2', '/service/2')

@service2.get(accept_catch_all=False)
def svc2_get(request):
    return {'test': 'succeeded'}




DEFAULT_VIEW_SETTINGS = {
    'renderer': 'jsonp',        # default views' renderer
    'help': True,               # enable 'help' by default
    'accept_catch_all': True    # set 'custom_predicate' to True by default
}


# Default message if a request with query parameter `help` is processed
# and the request does not have defined a schema.
DEFAULT_HELP_MESSAGE = {'help': 'This API endpoint does not accept any '
                                + 'specific query parameters'}

# ---------------------
#   JSONP_SETTINGS
# ---------------------

"""
We do add a JSONP renderer at startup
(lovely.pyrest.__init__.add_custom_config()). The renderer requires a
parameter name. If this parameter is given in a request as query parameter the
JSONP renderer will be triggered.

Since we cannot retrieve the specified parameter name by default when we need
it (e.g. in lovely.pyrest.views.py) we need to store the parameter name
(set_jsonp_param_name() and get_jsonp_param_name()).
"""

# Default query parameter name for a JSONP request.
DEFAULT_JSONP_PARAM_NAME = 'callback'

JSONP_SETTINGS = {
    # 'param_name' can be specified in the '.ini' file with following
    # identifier.
    'param_name_ini': 'lovely.pyrest.jsonp.param_name',

    # Override this to set a custom query parameter name using the '.ini' file.
    # Use 'set_jsonp_param_name(...)' to override the value.
    'param_name': DEFAULT_JSONP_PARAM_NAME
}


def set_jsonp_param_name(settings):
    """
    Set the specified JSONP query parameter name.
    The 'param_name' is being set to 'param_name_ini' if given;
    DEFAULT_JSONP_PARAM_NAME otherwise.

    Call the function once at startup time (e.g. lovely.pyrest.__init__).
    @settings
        pyramid config settings.
    """
    JSONP_SETTINGS['param_name'] = settings.get(
        JSONP_SETTINGS['param_name_ini'],
        DEFAULT_JSONP_PARAM_NAME
    )


def get_jsonp_param_name():
    """
    Return the specified JSONP query parameter name.
    If the parameter is set in the '.ini' file, the custom parameter is
    used; otherwise the default parameter is used.
    """
    return JSONP_SETTINGS['param_name']



from lovely.pyrest.service import Service
"""
To show the different error handling in the following example,
a request should have set the query parameter 'callback=your_callback' only.
Therefore, the request does not fulfil all required parameters specificed by
the schema below and a validation error occures.

The response could then look similar as follows:
Response: 200 OK
Content-Type: application/javascript; charset=UTF-8
your_callback({"status": "error", "errors": [{"location": "query", "description": "Required field 'id' is missing"}], "http_status": 400})
"""

svc_error = Service(name="svc_error", path='/svc/error')


schema = {
    "query": {
        "type": "object",
        "properties": {
            "id": {
                "type": "string"
            }
        }
    }
}

@svc_error.get(schema=schema)
def get(request):
    return {'title': 'Hello World!'}



from docutils import nodes
from docutils.parsers.rst import Directive
from lovely.pyrest.service import get_services
from importlib import import_module
from lovely.pyrest.sphinx.helpers import create_node, trim
from lovely.pyrest.sphinx.schema import SchemaDirective
from lovely.pyrest.sphinx.validators import ValidatorsDirective


def convert_to_list(value):
    if value is None:
        return []
    return [v.strip() for v in value.split(',')]


class ServiceDirective(Directive):
    """ The Service directive renders all services defined in a module """

    has_content = True
    required_arguments = 1

    option_spec = {'services': convert_to_list}

    def run(self):
        module = self.arguments[0]
        import_module(module)
        rendered = []
        settings = self.state.document.settings
        # fetch all services
        services = get_services()
        # if the services option is set render only the named services
        names = self.options.get('services')
        if names:
            services = [s for s in services if s.name in names]
        for service in services:
            service_id = "service_%d" % settings.env.new_serialno('service')
            rendered.append(ServiceDirective.render(service, service_id))
        return rendered

    @staticmethod
    def render(service, service_id):
        service_node = nodes.section(ids=[service_id])
        title = "%s service - %s" % (service.name, service.path)
        service_node += nodes.title(text=title)

        if service.description is not None:
            service_node += create_node(trim(service.description))

        for method, view, args in service.definitions:
            method_id = "%s_%s" % (service_id, method)
            method_node = nodes.section(ids=[method_id])
            method_node += nodes.title(text=method)

            # Render Accept Header documentation
            acceptables = service.acceptables(method)
            if len(acceptables) > 0:
                accept = 'Accept: ' + ', '.join(acceptables)
                method_node += create_node(accept)

            # Render Content-Type Header documentation
            content_types = service.content_types(method)
            if len(content_types) > 0:
                ct = 'Content-Types: ' + ', '.join(content_types)
                method_node += create_node(ct)

            if 'schema' in args:
                method_node += SchemaDirective.render_schema(service,
                                                             method,
                                                             service_id)

            if 'validators' in args:
                method_node += ValidatorsDirective.render(service,
                                                          method,
                                                          service_id)
            service_node += method_node
        return service_node



from pprint import pprint
import unittest
import doctest
import json
import sys
from pyramid.request import Request
from lovely.pyrest.errors import Errors
from docutils.core import publish_from_doctree, publish_doctree


def print_json(js):
    try:
        d = json.loads(js)
    except ValueError:
        print >> sys.stderr, js
        raise
    print(json.dumps(d, indent=4, sort_keys=True))


def render_doc_node(node, writer_name='pseudoxml'):
    """ Renderers a docutils node """
    # Create an empty document
    doc = publish_doctree("")
    # append the node
    doc.children.append(node)
    # publish the document and return the output
    return publish_from_doctree(doc, writer_name=writer_name)


def create_request():
    request = Request({})
    request.errors = Errors()
    return request


def setUp(test):
    test.globs['pprint'] = pprint
    test.globs['create_request'] = create_request
    test.globs['print_json'] = print_json
    test.globs['render_doc_node'] = render_doc_node


def tearDown(test):
    pass


def create_suite(testfile, layer=None, level=None,
                 tearDown=tearDown, setUp=setUp, cls=doctest.DocFileSuite):
    suite = cls(
        testfile, tearDown=tearDown, setUp=setUp,
        optionflags=doctest.NORMALIZE_WHITESPACE | doctest.ELLIPSIS)
    if layer:
        suite.layer = layer
    if level:
        suite.level = level
    return suite


def test_suite():
    s = unittest.TestSuite((
        create_suite('service.txt'),
        create_suite('validation.txt'),
        create_suite('predicates.txt'),
        create_suite('views.txt'),
        create_suite('rest.txt'),
        create_suite('__init__.txt'),
        create_suite('itests/service_definition.txt'),
        create_suite('itests/accept_catch_all.txt'),
        create_suite('sphinx/schema.txt'),
        create_suite('sphinx/validators.txt'),
        create_suite('sphinx/service.txt'),
    ))
    return s



from lovely.pyrest.sphinx.schema import SchemaDirective
from lovely.pyrest.sphinx.validators import ValidatorsDirective
from lovely.pyrest.sphinx.service import ServiceDirective


def setup(sphinx):
    sphinx.add_directive('service', ServiceDirective)
    sphinx.add_directive('schema', SchemaDirective)
    sphinx.add_directive('validators', ValidatorsDirective)
